@@ -36,15 +36,14 @@ class Client(object):
                 return value
 
     def get_multi(self, keys):
-        values = []
-        for key in keys:
+        d = {}
+        if keys:
             for server in self.servers:
-                value = server.get(key)
-                if value is not None:
-                    values.append((key, value))
+                d.update(server.get_multi(keys))
+                keys = [_ for _ in keys if not _ in d]
+                if not keys:
                     break
-
-        return dict(values)
+        return d
 
     def set(self, key, value, time=100):
         returns = []
@@ -55,8 +54,8 @@ class Client(object):
 
     def set_multi(self, mappings, time=100):
         returns = []
-        for key, value in mappings.iteritems():
-            returns.append(self.set(key, value, time))
+        for server in self.servers:
+            returns.append(server.set_multi(mappings, time))
 
         return all(returns)
 
@@ -126,7 +125,10 @@ class Server(object):
     # All structures will be appended to HEADER_STRUCT
     COMMANDS = {
         'get': {'command': 0x00, 'struct': '%ds'},
+        'getk': {'command': 0x0C, 'struct': '%ds'},
+        'getkq': {'command': 0x0D, 'struct': '%ds'},
         'set': {'command': 0x01, 'struct': 'LL%ds%ds'},
+        'setq': {'command': 0x11, 'struct': 'LL%ds%ds'},
         'add': {'command': 0x02, 'struct': 'LL%ds%ds'},
         'replace': {'command': 0x03, 'struct': 'LL%ds%ds'},
         'delete': {'command': 0x04, 'struct': '%ds'},
@@ -135,7 +137,7 @@ class Server(object):
         'flush': {'command': 0x08, 'struct': 'I'},
         'stat': {'command': 0x10},
         'auth_negotiation': {'command': 0x20},
-        'auth_request': {'command': 0x21, 'struct': '%ds%ds'}
+        'auth_request': {'command': 0x21, 'struct': '%ds%ds'},
     }
 
     STATUS = {
@@ -153,6 +155,8 @@ class Server(object):
         'compressed': 1 << 3
     }
 
+    COMPRESSION_THRESOLD = 128
+
     def __init__(self, server, username=None, password=None):
         self.server = server
         self.authenticated = False
@@ -266,8 +270,10 @@ class Server(object):
             flags |= self.FLAGS['pickle']
             value = dumps(value)
 
-        value = zlib.compress(value)
-        flags |= self.FLAGS['compressed']
+        if len(value) > self.COMPRESSION_THRESOLD:
+            value = zlib.compress(value)
+            flags |= self.FLAGS['compressed']
+
         return (flags, value)
 
     def deserialize(self, value, flags):
@@ -310,6 +316,41 @@ class Server(object):
 
         return self.deserialize(value, flags)
 
+    def get_multi(self, keys):
+        # pipeline N-1 getkq requests, followed by a regular getk to uncork the
+        # server
+        keys, last = keys[:-1], keys[-1]
+        msg = ''.join([
+            struct.pack(self.HEADER_STRUCT + \
+                self.COMMANDS['getkq']['struct'] % (len(key)),
+                self.MAGIC['request'],
+                self.COMMANDS['getkq']['command'],
+                len(key), 0, 0, 0, len(key), 0, 0, key)
+            for key in keys])
+        msg += struct.pack(self.HEADER_STRUCT + \
+            self.COMMANDS['getk']['struct'] % (len(last)),
+            self.MAGIC['request'],
+            self.COMMANDS['getk']['command'],
+            len(last), 0, 0, 0, len(last), 0, 0, last)
+
+        self.connection.send(msg)
+
+        d = {}
+        opcode = -1
+        while opcode != self.COMMANDS['getk']['command']:
+            (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
+                cas, extra_content) = self._get_response()
+
+            if status == self.STATUS['success']:
+                flags, key, value = struct.unpack('!L%ds%ds' %
+                    (keylen, bodylen - keylen - 4), extra_content)
+                d[key] = self.deserialize(value, flags)
+            elif status != self.STATUS['key_not_found']:
+                raise MemcachedException('Code: %d Message: %s' % (status,
+                    extra_content))
+
+        return d
+
     def _set_add_replace(self, command, key, value, time):
         logger.info('Setting/adding/replacing key %s.' % key)
         flags, value = self.serialize(value)
@@ -344,6 +385,43 @@ class Server(object):
     def replace(self, key, value, time):
         return self._set_add_replace('replace', key, value, time)
 
+    def set_multi(self, mappings, time=100):
+        mappings = mappings.items()
+        mappings, last = mappings[:-1], mappings[-1]
+        msg = []
+        for key, value in mappings:
+            flags, value = self.serialize(value)
+            m = struct.pack(self.HEADER_STRUCT + \
+                self.COMMANDS['setq']['struct'] % (len(key), len(value)),
+                self.MAGIC['request'],
+                self.COMMANDS['setq']['command'],
+                len(key),
+                8, 0, 0, len(key) + len(value) + 8, 0, 0, flags, time, key, value)
+            msg.append(m)
+
+        key, value = last
+        flags, value = self.serialize(value)
+        msg.append(struct.pack(self.HEADER_STRUCT + \
+            self.COMMANDS['set']['struct'] % (len(key), len(value)),
+            self.MAGIC['request'],
+            self.COMMANDS['set']['command'],
+            len(key),
+            8, 0, 0, len(key) + len(value) + 8, 0, 0, flags, time, key, value))
+
+        msg = ''.join(msg)
+
+        self.connection.send(msg)
+
+        opcode = -1
+        retval = True
+        while opcode != self.COMMANDS['set']['command']:
+            (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
+                cas, extra_content) = self._get_response()
+            if status != self.STATUS['success']:
+                retval = False
+
+        return retval
+
     def _incr_decr(self, command, key, value, default, time):
         self.connection.send(struct.pack(self.HEADER_STRUCT + \
             self.COMMANDS[command]['struct'] % len(key),
